/* cfft.c - functions for reading Chilbolton FFT data */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "cfft.h"

int cfft_error = CFFT_OK;
char cfft_verbose = 0;

int
cfft_open(char *filename, CFFTData *data)
{
  __CFFTHeader header;
  __CFFTOldHeader oldheader;
  FILE *file;
  CFFTFloat *spectra;
  double *rawspectra;
  CFFTFloat *range, *velocity;
  size_t nread;
  unsigned long npoints;
  int i;
  static const char out_of_memory[] = "CFFT Error: out of memory\n";

  file = fopen(filename, "r");
  if (!file) {
    if (cfft_verbose) {
      fprintf(stderr, "CFFT Error: failed to read %s\n", filename);
    }
    return (cfft_error = CFFT_FILE_ERROR);
  }

  nread = fread(&header, sizeof(__CFFTHeader), 1, file);
  if (nread < 1) {
    if (cfft_verbose) {
      fprintf(stderr, "CFFT Error: %s too short\n", filename);
    }
    return (cfft_error = CFFT_FILE_ERROR);
  }

  /* First check to see if we have the old data format... */
  if (header.file_header_fields == 250
      && header.profile_header_fields == 256) {
    /* Yes we do. - the actual data is stored as doubles */
    rewind(file);
    nread = fread(&oldheader, sizeof(__CFFTOldHeader), 1, file);

    /* Copy header information into the CFFTData structure */
    /* Some of the fields have not been set properly in early data */
    if ((data->divfactor = oldheader.divfactor) == 0) {
      data->divfactor = 2;
    }
    if ((data->delayclocks = oldheader.delayclocks) == 0) {
      data->delayclocks = 2;
    }
    if ((data->interval = oldheader.interval) == 0) {
      data->interval = 600;
    }
    data->ngates = oldheader.ngates;
    data->nbins = oldheader.nbins;
    data->naverages = oldheader.naverages;
    data->offset1 = oldheader.offset1;
    data->offset2 = oldheader.offset2;
    data->file_header_size = sizeof(__CFFTOldHeader);
    data->profile_header_size = 0;
    data->sizeof_real = sizeof(double);

    npoints = oldheader.ngates * oldheader.nbins;
  }
  else {
    /* New header format - and the data is stored as floats */
    data->divfactor = header.divfactor;
    data->delayclocks = header.delayclocks;
    data->interval = header.interval;
    data->ngates = header.ngates;
    data->nbins = header.nbins;
    data->naverages = header.naverages;
    data->offset1 = header.offset1;
    data->offset2 = header.offset2;
    data->file_header_size = header.file_header_fields * sizeof(int);
    data->profile_header_size = header.profile_header_fields * sizeof(int);
    data->sizeof_real = sizeof(float);

    npoints = header.ngates * header.nbins;

    /* Also read through any extra header fields that we don't yet understand */
    if (data->profile_header_size > sizeof(__CFFTHeader)) {
      fseek(file, data->profile_header_size - sizeof(__CFFTHeader), SEEK_CUR);
    }
  }

  range = malloc(sizeof(CFFTFloat) * data->ngates);
  velocity = malloc(sizeof(CFFTFloat) * data->nbins);
  spectra = malloc(sizeof(CFFTFloat) * npoints);

  if (!range || !velocity || !spectra) {
    if (cfft_verbose) {
      fprintf(stderr, "Trying to allocate real arrays with lengths of %d, %d and %ld: ",
	      data->ngates, data->nbins, npoints);
      fprintf(stderr, out_of_memory);
    }
    return (cfft_error = CFFT_MEMORY_ERROR);
  }

  for (i = 0; i < data->ngates; ++i) {
    range[i] = (i*data->divfactor+data->delayclocks)
      * 0.03 + data->range_offset*0.001;
  }

  if (data->flags & CFFT_REVERSE_V) {
    for (i = 0; i < data->nbins; ++i) {
      velocity[i] = - ((2*i - data->nbins) * 5.0) / (data->nbins);
    }
  }
  else {
    for (i = 0; i < data->nbins; ++i) {
      velocity[i] = ((2*i - data->nbins) * 5.0) / (data->nbins);
    }
  }

#ifdef CFFT_USE_FLOATS
  if (data->sizeof_real == sizeof(double)) {
    /* Spectra will be loaded into data->rawspectra as doubles
     * and copied into data->spectra as floats */
    rawspectra = malloc(sizeof(double) * npoints);
    if (!rawspectra) {
      if (cfft_verbose) {
	fprintf(stderr, "Trying to allocate double array with length %ld: ",
		npoints);
	fprintf(stderr, out_of_memory);
      }
      free(spectra);free(velocity);free(spectra);
      return (cfft_error = CFFT_MEMORY_ERROR);
    }
  }
  else {
    rawspectra = NULL;
  }
#else
  if (data->sizeof_real == sizeof(float)) {
    rawspectra = NULL;
  }
  else {
    free(spectra);free(velocity);free(spectra);
    return (cfft_error = "Sorry, I don't convert floats back to doubles\n");
  }
#endif

  /* Get the time from the appropriate raw file, if it exists */
  cfft_get_time(data, filename);

  /* Copy data into CFFTData structure */
  data->spectra = spectra;
  data->rawspectra = rawspectra;
  data->range = range;
  data->velocity = velocity;
  data->file = file;
  data->_nallocated = npoints;
  data->profile = -1;

  /* Set default processing parameters */
  data->noisegates = 50;
  data->noisefactor = 3.0;
  data->spikebleed = 1;

  return (cfft_error = CFFT_OK);
}

int
cfft_count_profiles(CFFTData *data)
{
  int ifile = fileno(data->file);
  struct stat fdata;
  int nprofiles = 0;

  if (fstat(ifile, &fdata)) {
    cfft_error = CFFT_FILE_ERROR;
    return -1;
  }
  nprofiles = (fdata.st_size - data->file_header_size)
    / (data->nbins * data->ngates * data->sizeof_real
       - data->profile_header_size);
  return nprofiles;
}

int
cfft_close(CFFTData *data)
{
  int status;
  if (data->_nallocated && data->spectra) {
    free(data->spectra);
    data->spectra = NULL;
  }
  if (data->rawspectra) {
    free(data->rawspectra);
    data->rawspectra = NULL;
  }
  if (data->range) {
    free(data->range);
    data->range = NULL;
  }
  if (data->velocity) {
    free(data->velocity);
    data->range = NULL;
  }
  status = fclose(data->file);
  if (status) {
    cfft_error = CFFT_FILE_ERROR;
  }
  else {
    cfft_error = CFFT_OK;
  }
  return status;
}

int
cfft_next_profile(CFFTData *data)
{
  size_t nread = 0;
  unsigned long npoints = data->nbins * data->ngates;
  CFFTFloat *spectra = data->spectra;
#ifdef CFFT_USE_FLOATS
  double *rawspectra = data->rawspectra;
  int i;
#endif

  /* Read header, if there is one */
  if (data->profile_header_size >= 6*sizeof(int)) {
    int tmpint, minutes_since_midnight, centiseconds;
    fread(&tmpint, sizeof(int), 1, data->file);
    fread(&minutes_since_midnight, sizeof(int), 1, data->file);
    fread(&centiseconds, sizeof(int), 1, data->file);
    fread(&tmpint, sizeof(int), 1, data->file);
    data->year = tmpint;
    fread(&tmpint, sizeof(int), 1, data->file);
    data->month = tmpint;
    fread(&tmpint, sizeof(int), 1, data->file);
    data->day = tmpint;
    data->hour = minutes_since_midnight / 60;
    data->minute = minutes_since_midnight - (data->hour * 60);
    data->second = centiseconds / 100;
    nread = 6;
  }

  /* Skip any unread header elements */
  if (data->profile_header_size > nread * sizeof(int)) {
    fseek(data->file, data->profile_header_size - nread * sizeof(int),
	  SEEK_CUR);
  }

  /* Read the actual data */
#ifdef CFFT_USE_FLOATS
  if (data->sizeof_real == sizeof(CFFTFloat)) {
    nread = fread(spectra, data->sizeof_real, npoints, data->file);
  }
  else {
#endif
    nread = fread(rawspectra, data->sizeof_real, npoints, data->file);
#ifdef CFFT_USE_FLOATS
  }
#endif

  if (nread < npoints) {
    if (feof(data->file)) {
      return (cfft_error = CFFT_EOF);
    }
    else {
      return (cfft_error = CFFT_FILE_ERROR);
    }
  }
#ifdef CFFT_USE_FLOATS
  if (rawspectra) {
    for (i = 0; i < npoints; ++i) {
      spectra[i] = rawspectra[i];
    }
  }
#endif

  ++(data->profile);

  if (data->profile_header_size < 6*sizeof(int)) {
    cfft_sync_time(data);
  }

  return (cfft_error = CFFT_OK);
}

/* Jump to particular profile and then run cfft_next_profile().
 * If profile is negative, start from the back of the file. */
int
cfft_read_profile(CFFTData *data, int profile)
{
  if (profile < 0) {
    int nprofiles = cfft_count_profiles(data);
    if (nprofiles < 1) {
      return (cfft_error = CFFT_FILE_ERROR);
    }
    else {
      long offset = (nprofiles + profile - data->profile - 1)
	* (data->ngates * data->nbins * data->sizeof_real
	   + data->profile_header_size);
      if (fseek(data->file, offset, SEEK_CUR)) {
	return (cfft_error = CFFT_FILE_ERROR);
      }
      data->profile = nprofiles + profile - 1;
    }
  }
  else {
    long offset = (profile - data->profile - 1)
      * (data->ngates * data->nbins * data->sizeof_real 
	 + data->profile_header_size);
    if (fseek(data->file, offset, SEEK_CUR)) {
      return (cfft_error = CFFT_FILE_ERROR);
    }
    data->profile = profile - 1;
  }

  return cfft_next_profile(data);
}
